import { Player, world, EntityComponentTypes, EquipmentSlot, ItemStack, ItemTypes, ItemLockMode, GameMode, ItemComponentTypes, EnchantmentType, system, BlockVolume } from '@minecraft/server';
import { ActionFormData } from '@minecraft/server-ui';
const overworld = world.getDimension('overworld');

const places = [
    {
        x: 10,
        y: 4,
        z: 401
    },  
    {
      x: 17,
      y: 4,
      z: 410
    },  
    {
      x: 4,
      y: 4,
      z: 413
    },  
    {
      x: 11,
      y: 4,
      z: 425
    },  
    {
      x: 28,
      y: 4,
      z: 402
    },  
    {
      x: 1,
      y: 4,
      z: 434
    },  
    {
      x: 31,
      y: 4,
      z: 418
    }  ,  
    {
      x: 23,
      y: 4,
      z: 423
    }  ,  
    {
      x: 29,
      y: 4,
      z: 431
    }  ,  
    {
      x: 14,
      y: 4,
      z: 436
    }  

];

const lobby = {
    x: 0,
    y: 67,
    z: 0
};

const center = {
    x: 19,
    y: -26,
    z: 419 
};

const refCorner = {
    x: 0,
    y: -52,
    z: 400
};

let player_places = places.map((p) => {
    p.ocupied_by = null;
    return p;
});

/** 
 * @param {Player} player 
 * @param {boolean} matchEnded 
 */
function spawnTeleport(player, matchEnded) {
    player._lastHitBy = null;
    player.setGameMode(GameMode.adventure);
    player.removeTag('spectating');
    player.removeTag('playing');
    player.removeTag('excluded');

    // world.sendMessage(`debug: tp to ${player.name} to lobby`);

    const invComp = player.getComponent(EntityComponentTypes.Inventory);
    const healthComp = player.getComponent(EntityComponentTypes.Health);
    const equipComp = player.getComponent(EntityComponentTypes.Equippable);
    invComp.container.clearAll();
    player.extinguishFire();

    try {
        player.getEffects().forEach((effect) => {
            if(effect.isValid()) player.removeEffect(effect.typeId);
    });
    } catch {}

    player.resetLevel();
    healthComp.resetToMaxValue();

    for (const slot in EquipmentSlot) {
        equipComp.setEquipment(slot, null);
    }
    
    player.teleport(lobby, {
        'facingLocation': {
            x: 1,
            y: 67,
            z: 0
        },
        'dimension': world.getDimension('overworld')
    });

    if (player.isOp()) {
        const item = new ItemStack('compass');
        item.nameTag = '§r§fManage Game';
        item.lockMode = ItemLockMode.slot;
        item.keepOnDeath = true;
        invComp.container.addItem(item);
    }

    const item = new ItemStack('enchanted_book');
    item.nameTag = '§r§fCosmethics';
    item.lockMode = ItemLockMode.slot;
    item.keepOnDeath = true;

    invComp.container.setItem(8, item);

    if (matchEnded) {
        invComp.container.addItem(new ItemStack('tadase:tada'));
    }

}

/** 
 * @param {Player} player 
 * @param {boolean} spectate 
 */
function gameTeleport(player, spectate = false) {
    const invComp = player.getComponent(EntityComponentTypes.Inventory);

    invComp.container.clearAll();
    player.camera.fade({
        'fadeColor': { red: 0, green: 0, blue: 0 },
        'fadeTime': {
            'fadeInTime': 0.4,
            'holdTime': 0.2,
            'fadeOutTime': 0.5
        }
    });

    if (spectate) {
        player.teleport(center, {
            'facingLocation': center
        });
    } else {

        let sucess = false;
        for (const i in player_places) {
            const place = player_places[i];
            if (place.ocupied_by) continue;
            player_places[i].ocupied_by = player.name;

            player.teleport({
                x: place.x + 0.5,
                y: place.y,
                z: place.z + 0.5
            },
                {
                    'dimension': overworld,
                    'keepVelocity': false,
                    'facingLocation': {
                        x: center.x,
                        y: place.y + 1,
                        z: center.z
                    }
                }
            );
            
            sucess = true;
            player.addTag('playing');

            break;
        }

        
        if (!sucess) {
            player.setGameMode(GameMode.spectator);
            player.addTag('spectating');
            player.addTag('excluded');

            gameTeleport(player, true);
        }

        

    }
}

function clearGameArea() {
    const overworld = world.getDimension('overworld');

    // world.sendMessage('debug: clear game area');
    let begin = -64;
    let end = begin + 3;
    const i = system.runInterval(() => {

        if (end > 14) {
            system.clearRun(i);
            world.structureManager.place('tadase:10_map1', overworld, refCorner);

            overworld.getEntities({
                'excludeTypes': ['minecraft:player', 'tadase:text', 'tadase:title'],
                location: center
            }).forEach((e) => {
                e.remove()
            });
            
        } else {
            overworld.fillBlocks(
                new BlockVolume({
                    x: center.x - 40,
                    y: begin,
                    z: center.z - 40
                },
                {
                    x: center.x + 40,
                    y: end,
                    z: center.z + 40
                }),
                'air'
            );    

            begin = end;
            end = begin + 3;
        }
    }, 2);
}

function restart() {

    let av = places.map((_, i) => i);
    player_places = []
    for (const i in av) {
        const random = Math.floor(Math.random() * av.length);
        const place = places[av[random]];
        place.ocupied_by = null;
        player_places.push(place);

        av = av.filter((i) => i !== av[random])
    }

}

function resetObjectives() {
    try {
        if(world.scoreboard.getObjective('kills')) world.scoreboard.removeObjective('kills');
    } catch (e) {
        console.warn('[UTILS] Cannot remove Objective.')
    }

    return world.scoreboard.addObjective('kills', '');
}


function updatePlayerCount() {
    const ent = overworld.getEntities({
        type: 'tadase:text',
        tags: ['pc']
    })[0];

    const players = world.getPlayers();
    if (!ent) return players.length;

    let newtag = `§${players.length > 1 ? 'a' : 'c'}(${players.length}/10) players`;

    if (players.length <= 1) {
        newtag += `\n2 players to start`;
    }

    if (players.length >= 10) {
        newtag += `\nFull`;
    }

    ent.nameTag = newtag;

    return players.length;
}

function removeBox(place = {}) {
        overworld.fillBlocks(new BlockVolume({
            x: place.x + 2,
            y: place.y,
            z: place.z + 2
        }, 
        {
            x: place.x - 1,
            y: place.y + 5,
            z: place.z -1
        }), 'air');
}

function removeUnusedBoxes() {
    let current = 0;
    const i = system.runInterval(() => {
        if (current < player_places.length) {
            const place = player_places[current];
                removeBox({
                    x: place.x,
                    y: place.y - 1,
                    z: place.z
                });
            
            current++;
        } else {
            system.clearRun(i);
        }
    });
}

/**
 * @param {Player} player 
 * @param {Object} title 
 * @param {Object} body 
 * @returns {boolean}
 */
function confirmationForm(player, title, body) {
    return new Promise((resolve, reject) => {
        new ActionFormData()
        .title(title)
        .body(body)
        .button({ translate: 'ui.generic.confirmation.ok' })
        .button({ translate: 'ui.generic.confirmation.no' })
        .show(player)
        .then((res) => {
            if (res.canceled || res.selection === 1) return resolve(false);
            if (res.selection === 0) return resolve(true);
        });
    });
}

export {
    spawnTeleport,
    gameTeleport,
    clearGameArea,
    restart,
    resetObjectives,
    updatePlayerCount,
    removeBox,
    removeUnusedBoxes,
    confirmationForm,
    lobby,
    center
}