import { world, GameMode, system, EntityComponentTypes, ItemStack, ItemTypes, ScoreboardObjective, BlockTypes, Player, Entity, DisplaySlotId, ObjectiveSortOrder, EffectTypes, TicksPerSecond, MolangVariableMap, EquipmentSlot, ItemComponentTypes, EntityDamageCause, ItemType, WorldAfterEvents } from '@minecraft/server';
import { spawnTeleport, gameTeleport, clearGameArea, restart, resetObjectives, lobby, updatePlayerCount, removeBox, removeUnusedBoxes, confirmationForm, center } from './Utils';
import { ActionFormData, ModalFormData } from '@minecraft/server-ui';
import items from './Items';

// wins 
// top 
// skull 
//  kills 
// name   

const overworld = world.getDimension('overworld');
let lastDead; 
let endTime = Date.now();
let started = false;
let firsPlayer = false;

const num_emojis = ['', '', '', '', ''];
const potionIds = [5, 7, 9, 12, 14, 17, 19, 21, 23, 25, 28, 31, 34, 36, 37, 40];

/**
 * @type {ScoreboardObjective} 
 */
let scores = world.scoreboard.getObjective('kills');

world.afterEvents.playerSpawn.subscribe((e) => {
    if (e.initialSpawn) {
        if (started) {
            e.player.removeTag('playing');
            e.player.addTag('spectating');
            e.player.addTag('excluded');
            gameTeleport(e.player, true);

        } else {
            if (!firsPlayer) {
                firsPlayer = true;
                clearGameArea();
            }

            const pcount = updatePlayerCount();
            if (pcount === 10) {
                world.sendMessage({
                    translate: 'lobby.party_full'
                });
            }

            e.player.removeTag('spectating');
            e.player.removeTag('playing');
            e.player.removeTag('excluded');
            spawnTeleport(e.player);
            e.player.sendMessage(' -------------------------- \n§lWelcome to    \nDeveloped by §dTadaseCake.§f\n -------------------------- ');
            e.player.sendMessage({ translate: 'info.discord'});
            e.player.sendMessage(' §bdiscord.gg/UHCKr8yCJ5');
        }

    } else {
        if (!started) {
            return spawnTeleport(e.player);
        }

        const players = world.getPlayers({
            tags: ['playing']
        });

        if (e.player.hasTag('spectating')) {
            e.player._validDeath = true;
            e.player.setGameMode(GameMode.spectator);
            e.player.onScreenDisplay.setTitle('');
            gameTeleport(e.player, true);
            players.forEach((player) => {
                player.playSound('note.bass');
            });
        }

        if (players.length <= 1 || !players.length) {
            started = false;
            const winner = players[0] || e.player;

            system.run(() => {
                world.setDynamicProperty('gameRunning', false);
                world.sendMessage({
                    'translate': 'game.winner',
                    with: [winner.name]
                });

                winner.addEffect('levitation', TicksPerSecond * 1, { amplifier: 3, showParticles: false });
                winner.addEffect('slow_falling', TicksPerSecond * 5, { amplifier: 98, showParticles: false });
                winner.addEffect('levitation', TicksPerSecond * 5, { amplifier: 1, showParticles: false });
                winner.addEffect('regeneration', TicksPerSecond * 5, { amplifier: 98, showParticles: false });

                overworld.spawnParticle('tadase:confetti', winner.location);

                restart();
                overworld.runCommand('difficulty 0');
                overworld.runCommand('gamerule pvp false');

                world.sendMessage({
                    translate: 'game.finished'
                });
                const killsObjective = world.scoreboard.getObjective('kills');

                const all = world.getPlayers();
                all.forEach((player) => {
                    player.playSound('random.levelup');
                    player.playSound('fireworks.launch');
                    player.playSound('mob.ghast.fireball');
                });

                const text = all.filter((player) => !player.hasTag('excluded')).map((player) => ({ name: player.name, score: killsObjective.getScore(player) }))
                    .sort((a, b) => b.score - a.score).map((p, ind) => `§b${ind + 1})§f ${p.name} ----- ${p.score}`).join('\n');
                world.sendMessage(text);
                world.sendMessage({
                    translate: 'lobby.tp'
                });
                system.runTimeout(() => {
                    all.forEach((player) => {
                        console.log('teleporting');
                        spawnTeleport(player, true);
                    });

                    overworld.spawnParticle('tadase:confetti', lobby);
                    scores = resetObjectives();

                    const ent = overworld.getEntities({
                        type: 'tadase:text',
                        tags: ['lb']
                    })[0];
                    overworld.runCommand('effect @a clear');
                    clearGameArea();
                    endTime = Date.now();

                    if (ent) ent.nameTag = '\n' + text;
                }, TicksPerSecond * 4);


            });

        }
    }
});

world.afterEvents.itemUse.subscribe((e) => {
    if (started) return;
    if (e.itemStack.typeId === 'minecraft:compass' && e.source.isOp()) {

        let players = world.getPlayers();
        new ActionFormData()
            .title({
                translate: 'ui.manager.title'
            })
            .body('                 \n\n                     \n\n              \n\n                 §lby TadaseCake')
            .button({ translate: 'ui.manager.settings' })
            .button({
                translate: 'ui.manager.start',
                with: [`\n§r§${players.length > 1 ? 'a' : 'c'}(${players.length}/10)`]
            })
            .show(e.source)
            .then((res) => {
                if (res.canceled || started) return;

                if (res.selection === 0) {
                    const prop = world.getDynamicProperty('config');
                    let config = prop ? JSON.parse(prop) : { healOnKill: 0 };
                    
                    new ModalFormData()
                        .title({ translate: 'ui.manager.settings' })
                        .slider({ translate: 'ui.settings.heal_on_kill' }, 0, 10, 1, config.healOnKill)
                        .submitButton({ translate: 'ui.settings.submit' })
                        .show(e.source).then((res) => {

                            if (res.canceled) return;
                            config.healOnKill = res.formValues[0] || 0;
                            world.setDynamicProperty('config', JSON.stringify(config));
                            e.source.sendMessage({
                                translate: 'settings.updated'
                            });
                            e.source.playSound('block.enchanting_table.use');

                        });

                } else if (res.selection === 1) {
                    const secs = (Date.now() - endTime) / 1000;
                    if (secs < 6) {
                        return e.source.sendMessage({
                            'translate': 'lobby.alert.start_cooldown',
                            with: [Math.round(6 - secs) + '.' + Math.round(secs % 60)]
                        });
                    }

                    if (players.length <= 0) {
                        return e.source.sendMessage({
                            translate: 'lobby.alert.enough_players'
                        });
                    }

                    world.sendMessage({
                        translate: 'game.tp'
                    });

                    players.forEach((player) => {
                        player.playSound('portal.travel');
                        gameTeleport(player, false);
                    });

                    restart();
                    scores = resetObjectives();
                    world.scoreboard.setObjectiveAtDisplaySlot(DisplaySlotId.Sidebar, {
                        'objective': scores,
                        'sortOrder': ObjectiveSortOrder.Descending
                    });

                    world.sendMessage({
                        translate: 'lobby.alert.starting'
                    });

                    system.runTimeout(() => {
  
                        let count = 10;
                        world.sendMessage({
                            translate: 'game.starting',
                            with: [count + '']
                        });
                        const i = system.runInterval(() => {

                            if (system.currentTick % 20 === 0) {

                                if (count == 0) {
                                    system.clearRun(i);
                                    world.setDynamicProperty('gameRunning', true);
                                    world.sendMessage({
                                        translate: 'game.started'
                                    });

                                    overworld.runCommand('difficulty hard');
                                    overworld.runCommand('gamerule pvp true');
                                    started = true;

                                    players.forEach((player) => {
                                        if (!player.hasTag('playing')) return;
                                        overworld.getBlock({
                                            x: player.location.x,
                                            y: player.location.y - 1,
                                            z: player.location.z
                                        }).setType('air');

                                        player.addEffect('regeneration', TicksPerSecond * 3, {
                                            'amplifier': 5,
                                            'showParticles': false,
                                        });

                                        player.setGameMode(GameMode.survival);
                                        world.scoreboard.getObjective('kills').setScore(player, 0);

                                    });

                                    removeUnusedBoxes();
                                    return;
                                }

                                players.forEach((player) => {
                                    if (count > 5) {
                                        player.onScreenDisplay.setActionBar({
                                            translate: 'game.actionbar.count',
                                            with: [count + '']
                                        });
                                    } else {
                                        if (count === 5) {
                                            world.sendMessage({
                                                translate: 'game.starting',
                                                with: [count + '']
                                            });
                                        }

                                        player.onScreenDisplay.setTitle(num_emojis[count - 1], {
                                            subtitle: {
                                                translate: 'game.subtitle.starting'
                                            },
                                            stayDuration: 20,
                                            fadeInDuration: 5,
                                            fadeOutDuration: 10
                                        });
    
                                        player.playSound('note.bit', {
                                            pitch: 1 - (count / 10)
                                        });
                                    }

                                })
                                count--;
                            }
                        });
                    })



                }
            });
    }

    if (e.itemStack.typeId === 'minecraft:enchanted_book') {
        new ActionFormData()
            .title({
                translate: 'ui.title.cosmethics'
            })
            .button({ translate: 'generic.coming_soon' })
            .show(e.source);
    }

    if (e.itemStack.typeId === 'tadase:tada') {
        const cooldownComp = e.itemStack.getComponent(ItemComponentTypes.Cooldown);
        const playerRemain = cooldownComp.getCooldownTicksRemaining(e.source);
        if (playerRemain) return;

        const viewdir = e.source.getViewDirection();
        const molang = new MolangVariableMap();
        molang.setFloat('variable.rotation_x', viewdir.x * 20);
        molang.setFloat('variable.rotation_z', viewdir.z * 20);

        e.source.dimension.spawnParticle('tadase:manual_confetti', {
            x: e.source.location.x,
            y: e.source.location.y + 0.2,
            z: e.source.location.z
        }, molang);

        const equipComp = e.source.getComponent(EntityComponentTypes.Equippable);

        let hand = equipComp.getEquipment(EquipmentSlot.Mainhand);
        if (hand.amount == 1) {
            hand = null;
        } else {
            hand.amount--;
        }
        overworld.playSound('firework.launch', e.source.location, {
            'volume': 0.5,
            'pitch': 2
        });
        overworld.playSound('block.cartography_table.use', e.source.location, {
            'volume': 0.4,
            'pitch': 2
        });

        equipComp.setEquipment(EquipmentSlot.Mainhand, hand);
        cooldownComp.startCooldown(e.source);
    }
});

world.afterEvents.entityDie.subscribe((e) => {
    if (e.deadEntity.typeId !== 'minecraft:player') return;
    if (started) {
        if (e.deadEntity.hasTag('playing')) {
            e.deadEntity.addTag('spectating');
            e.deadEntity.removeTag('playing');
            /**
             * @type {Player}
             */
            const damager = e.damageSource.damagingEntity || e.deadEntity._lastHitBy;
            if (damager?.typeId === 'minecraft:player') {
                if (!e.damageSource.damagingEntity && e.deadEntity._lastHitBy) {

                    world.sendMessage({
                        translate: 'game.death.indirect',
                        with: [e.deadEntity.name, damager.name]
                    });
                }

                world.scoreboard.getObjective('kills')?.addScore(damager, 1);
                damager.playSound('firework.large_blast');
                damager.onScreenDisplay.setActionBar(' +1');  

                const prop = world.getDynamicProperty('config');
                let config = prop ? JSON.parse(prop) :
                    {
                        healOnKill: 0
                    };

                if (config.healOnKill) {
                    const healthComp = damager.getComponent(EntityComponentTypes.Health);
                    healthComp.setCurrentValue(healthComp.currentValue + config.healOnKill * 2);
                    overworld.spawnParticle('tadase:heal', damager.location);
                    damager.playSound('block.enchanting_table.use');
                }
            } else if (damager) {
                world.sendMessage({
                    translate: 'game.death.entity'
                });
            }
        }

    } else {
        if (e.deadEntity.hasTag('playing')) {
            e.deadEntity.addTag('spectating');
            e.deadEntity.removeTag('playing');
        }
    }

});

world.afterEvents.entityHitEntity.subscribe((e) => {
    if (!started) return;
    if (e.damagingEntity.typeId === 'minecraft:player' && e.hitEntity.typeId === 'minecraft:player') {
        const equipComp = e.damagingEntity.getComponent(EntityComponentTypes.Equippable);
        const hand = equipComp.getEquipment(EquipmentSlot.Mainhand);
        if (hand?.typeId === 'tadase:knockback_stick') {
            const view = e.damagingEntity.getViewDirection();
            e.hitEntity.applyKnockback(view.x, view.z, 3, 0);
        }

    }
});

world.afterEvents.worldInitialize.subscribe((e) => {
    try {
        world.scoreboard.removeObjective('kills');
    } catch (e) { }

    if (world.getDynamicProperty('gameRunning')) {
    }
});

world.afterEvents.entitySpawn.subscribe((e) => {
    if (e.entity.typeId === 'minecraft:ender_dragon') {
        system.runTimeout(() => {
            e.entity.remove();
        }, TicksPerSecond * 3);
    }
});

world.beforeEvents.playerPlaceBlock.subscribe((e) => {
    if ((Math.sqrt(Math.pow(e.block.location.x - center.x, 2) + Math.pow(e.block.location.z - center.z, 2)) >= 40 ||  e.block.location.y - center.y >= 40)&& e.player.hasTag('playing')) {
        e.cancel = true;

        system.run(() => {
            e.player.sendMessage({
                translate: 'game.alert.block_out_area'
            });
            e.player.playSound('note.bass');
        });

    }
});


world.afterEvents.playerLeave.subscribe((e) => {
    updatePlayerCount();
});

world.afterEvents.entityHurt.subscribe((e) => {
    if (e.hurtEntity.typeId === 'minecraft:player' && e.damageSource.cause === EntityDamageCause.void) {
        e.hurtEntity.applyDamage(200, {
            'cause': EntityDamageCause.void
        });
    }

    if (e.hurtEntity.typeId === 'minecraft:player' && e.damageSource.damagingEntity?.typeId === 'minecraft:player') {
        if (e.hurtEntity.name === e.damageSource.damagingEntity.name) return;
        e.hurtEntity._lastHitBy = e.damageSource.damagingEntity;
    }

});

system.runInterval(() => {
    if (!started) {
        if (system.currentTick % (TicksPerSecond * 180) === 0) {
            world.sendMessage({ translate: 'info.discord'});
            world.sendMessage(' §bdiscord.gg/UHCKr8yCJ5');
        }

        return;
    };
    const players = world.getPlayers({
        tags: ['playing']
    });
    if (system.currentTick % 100 === 0) {
        players.forEach((player) => {
            const random = items[Math.floor(Math.random() * items.length)];
            if (random === 'minecraft:potion' || random === 'minecraft:splash_potion') {
                player.runCommand(`give @s ${random} 1 ${potionIds[Math.floor(Math.random() * potionIds.length)]}`);
            } else {
                const giveItem = new ItemStack(random);
                player.getComponent(EntityComponentTypes.Inventory).container.addItem(giveItem);
            }

            player.playSound('random.pop');
        });
    }

    if (system.currentTick % 20 === 0) {
        const location = center;
        location.y = 0;
        const spectators = overworld.getEntities({
            tags: ['spectating'],
            type: 'minecraft:player',
            minDistance: 50,
            location
        });

        spectators.forEach((player) => {
            player.teleport(location);
            player.playSound('note.bass');
        });
    }

});